# =============================================================================
# Minet FileSystem Utilities
# =============================================================================
#
# Multiple helper functions related to reading and writing files.
#
import os
import gzip
import codecs
from threading import Lock
from os.path import basename, join
from ural import get_hostname, get_normalized_hostname


def read_potentially_gzipped_path(path, encoding='utf-8'):
    if path.endswith('.gz'):
        with open(path, 'rb') as f:
            raw_bytes = gzip.decompress(f.read())

        raw = raw_bytes.decode(encoding, errors='replace')
    else:
        with codecs.open(path, 'r', encoding=encoding, errors='replace') as f:
            raw = f.read()

    return raw


MAKEDIRS_LOCK = Lock()


def threadsafe_makedirs(path):
    with MAKEDIRS_LOCK:
        os.makedirs(path, exist_ok=True)


class FolderStrategy(object):
    def apply(self):
        raise NotImplementedError

    @staticmethod
    def from_name(name):
        if name == 'flat':
            return FlatFolderStrategy()

        if name == 'hostname':
            return HostnameFolderStrategy()

        if name == 'normalized-hostname':
            return NormalizedHostnameFolderStrategy()

        if name.startswith('prefix-'):
            length = name.split('prefix-')[-1]

            try:
                length = int(length)
            except ValueError:
                return None

            if length <= 0:
                return None

            return PrefixFolderStrategy(length)

        raise None


class FlatFolderStrategy(FolderStrategy):
    def apply(self, filename, **kwargs):
        return filename


class PrefixFolderStrategy(FolderStrategy):
    def __init__(self, length):
        self.length = length

    def apply(self, filename, **kwargs):
        base = basename(filename).split('.', 1)[0]

        return join(base[:self.length], filename)


class HostnameFolderStrategy(FolderStrategy):
    def apply(self, filename, url, **kwargs):
        hostname = get_hostname(url)

        if not hostname:
            hostname = 'unknown-host'

        return join(hostname, filename)


class NormalizedHostnameFolderStrategy(FolderStrategy):
    def apply(self, filename, url, **kwargs):
        hostname = get_normalized_hostname(
            url,
            normalize_amp=False,
            strip_lang_subdomains=True,
            infer_redirection=False
        )

        if not hostname:
            hostname = 'unknown-host'

        return join(hostname, filename)



# =============================================================================
# Minet CLI Utils
# =============================================================================
#
# Miscellaneous helpers used by the CLI tools.
#
import sys
import yaml
import platform
from glob import iglob
from os.path import join, expanduser, isfile, isdir
from collections import namedtuple
from tqdm import tqdm
from ebbe import noop

from minet.cli.exceptions import MissingColumnError
from minet.utils import fuzzy_int


def print_err(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def die(msg=None):
    if msg is not None:
        if not isinstance(msg, list):
            msg = [msg]

        for m in msg:
            print_err(m)

    sys.exit(1)


def safe_index(l, e):
    try:
        return l.index(e)
    except ValueError:
        return None


class LoadingBar(tqdm):
    def __init__(self, desc, stats=None, unit=None, unit_plural=None,
                 total=None, **kwargs):

        if unit is not None and total is None:
            if unit_plural is not None:
                unit = ' ' + unit_plural
            else:
                unit = ' ' + unit + 's'

        self.__stats = stats or {}

        if unit is not None:
            kwargs['unit'] = unit

        super().__init__(desc=desc, total=total, **kwargs)

    def update_total(self, total):
        self.total = total

    def update_stats(self, **kwargs):
        for key, value in kwargs.items():
            self.__stats[key] = value

        return self.set_postfix(**self.__stats)

    def inc(self, name):
        if name not in self.__stats:
            self.__stats[name] = 0

        self.__stats[name] += 1
        return self.update_stats()

    def print(self, *args, end='\n'):
        msg = ' '.join(str(arg) for arg in args)
        self.write(msg, file=sys.stderr, end=end)

    def die(self, msg):
        self.close()
        die(msg)


def acquire_cross_platform_stdout():

    # As per #254: stdout need to be wrapped so that windows get a correct csv
    # stream output
    if 'windows' in platform.system().lower():
        return open(
            sys.__stdout__.fileno(),
            mode=sys.__stdout__.mode,
            buffering=1,
            encoding=sys.__stdout__.encoding,
            errors=sys.__stdout__.errors,
            newline='',
            closefd=False
        )

    return sys.stdout


WorkerPayload = namedtuple(
    'WorkerPayload',
    ['row', 'headers', 'path', 'encoding', 'content', 'args']
)

REPORT_HEADERS = ['status', 'filename', 'encoding', 'mimetype']


def create_report_iterator(cli_args, reader, worker_args=None, on_irrelevant_row=noop):
    for col in REPORT_HEADERS:
        if col not in reader.pos:
            raise MissingColumnError(col)

    status_pos = reader.pos.status
    filename_pos = reader.pos.filename
    encoding_pos = reader.pos.encoding
    mimetype_pos = reader.pos.mimetype
    raw_content_pos = reader.pos.get('raw_contents')

    if raw_content_pos is None and not isdir(cli_args.input_dir):
        raise NotADirectoryError

    indexed_headers = reader.pos.as_dict()

    def generator():
        for row in reader:
            status = fuzzy_int(row[status_pos]) if row[status_pos] else None
            mimetype = row[mimetype_pos]
            filename = row[filename_pos]

            if status is None:
                on_irrelevant_row('no-status', row)
                continue

            if status != 200:
                on_irrelevant_row('invalid-status', row)
                continue

            if not filename:
                on_irrelevant_row('no-filename', row)
                continue

            if '/htm' not in mimetype:
                on_irrelevant_row('invalid-mimetype', row)
                continue

            if raw_content_pos is not None:
                yield WorkerPayload(
                    row=row,
                    headers=indexed_headers,
                    path=None,
                    encoding=row[encoding_pos],
                    content=row[raw_content_pos],
                    args=worker_args
                )

                continue

            path = join(cli_args.input_dir, filename)
            encoding = row[encoding_pos].strip() or 'utf-8'

            yield WorkerPayload(
                row=row,
                headers=indexed_headers,
                path=path,
                encoding=encoding,
                content=None,
                args=worker_args
            )

    return generator()


def create_glob_iterator(cli_args, worker_args):
    for p in iglob(cli_args.glob, recursive=True):
        yield WorkerPayload(
            row=None,
            headers=None,
            path=p,
            encoding='utf-8',
            content=None,
            args=worker_args
        )


def get_rcfile(rcfile_path=None):
    home = expanduser('~')

    files = [
        rcfile_path,
        '.minetrc',
        '.minetrc.yml',
        '.minetrc.yaml',
        '.minetrc.json',
        join(home, '.minetrc'),
        join(home, '.minetrc.yml'),
        join(home, '.minetrc.yaml'),
        join(home, '.minetrc.json')
    ]

    for p in files:
        if p is None or not isfile(p):
            continue

        with open(p, encoding='utf-8') as f:
            return yaml.safe_load(f)

    return None



# =============================================================================
# Minet Scraper Analysis
# =============================================================================
#
# Functions performing analysis of scraper definitions.
#
import ast
import soupsieve
from soupsieve import SelectorSyntaxError

from minet.scrape.utils import get_sel, get_iterator
from minet.scrape.constants import (
    PLURAL_MODIFIERS,
    BURROWING_KEYS,
    LEAF_KEYS,
    EXTRACTOR_NAMES
)
from minet.scrape.exceptions import (
    ScraperEvalSyntaxError,
    ScraperValidationConflictError,
    InvalidCSSSelectorError,
    ScraperValidationIrrelevantPluralModifierError,
    ScraperValidationMixedConcernError,
    ScraperValidationInvalidPluralModifierError,
    ScraperValidationInvalidExtractorError
)


class ScraperAnalysis(object):
    __slots__ = ('headers', 'plural', 'output_type')

    def __init__(self, headers=None, plural=False, output_type='scalar'):
        self.headers = headers
        self.plural = plural
        self.output_type = output_type

    def __eq__(self, other):
        return (
            self.plural == other.plural and
            self.output_type == other.output_type and
            set(self.headers) == set(other.headers)
        )

    def __repr__(self):
        return '<{name} headers={headers!r} plural={plural} output_type={output_type}>'.format(
            name=self.__class__.__name__,
            headers=self.headers,
            plural=self.plural,
            output_type=self.output_type
        )


def headers_from_definition(scraper):
    tabulate = scraper.get('tabulate')

    if tabulate is not None:
        if not isinstance(tabulate, dict):
            return None

        return tabulate.get('headers')

    fields = scraper.get('fields')

    if fields is None:
        return ['value']
    else:
        for node in fields.values():
            if 'fields' in node:
                return None

    return list(fields.keys())


def analyse(scraper):
    analysis = ScraperAnalysis()

    analysis.headers = headers_from_definition(scraper)

    iterator = get_iterator(scraper)

    if iterator is not None or 'iterator_eval' in scraper:
        analysis.plural = True
        analysis.output_type = 'list'

    if 'fields' in scraper:
        if analysis.plural:
            analysis.output_type = 'collection'
        else:
            analysis.output_type = 'dict'

    elif 'eval' in scraper:
        analysis.output_type = 'unknown'

    return analysis


ERRORS_PRIORITY = {
    InvalidCSSSelectorError: 0,
    ScraperEvalSyntaxError: 1,
    ScraperValidationConflictError: 2,
    ScraperValidationMixedConcernError: 3,
    ScraperValidationIrrelevantPluralModifierError: 4,
    ScraperValidationInvalidPluralModifierError: 5,
    ScraperValidationInvalidExtractorError: 6
}


def errors_sorting_key(error):
    return (tuple(error.path), ERRORS_PRIORITY[type(error)])


def validate(scraper):

    errors = []

    def recurse(node, path=[]):

        # Checking conflicts
        c = []

        if 'item' in node:
            c.append('item')
        if 'fields' in node:
            c.append('fields')
        if 'tabulate' in node:
            c.append('tabulate')

        if len(c) > 1:
            validation_error = ScraperValidationConflictError(
                path=path,
                keys=c
            )

            errors.append(validation_error)

        # Validating extractors
        if 'extract' in node and node['extract'] not in EXTRACTOR_NAMES:
            errors.append(ScraperValidationInvalidExtractorError(
                path=path,
                extractor=node['extract']
            ))

        # Validating selectors
        # NOTE: this has the beneficial side effect of precompiling selectors
        k, sel = get_sel(node, with_key=True)

        if sel is not None:
            try:
                soupsieve.compile(sel)
            except (SelectorSyntaxError, NotImplementedError) as e:
                errors.append(InvalidCSSSelectorError(path=path + [k], reason=e, expression=sel))

        k, iterator = get_iterator(node, with_key=True)

        if iterator is not None:
            try:
                soupsieve.compile(iterator)
            except (SelectorSyntaxError, NotImplementedError) as e:
                errors.append(InvalidCSSSelectorError(path=path + [k], reason=e, expression=iterator))

        # Checking plural modifiers
        if iterator is None and 'iterator_eval' not in node:
            for modifier in PLURAL_MODIFIERS:
                if modifier in node:
                    errors.append(ScraperValidationIrrelevantPluralModifierError(path=path, modifier=modifier))

        for modifier in PLURAL_MODIFIERS:
            if modifier not in node:
                continue

            if isinstance(node[modifier], bool):
                if 'fields' in node:
                    errors.append(ScraperValidationInvalidPluralModifierError(path=path, modifier=modifier))
            else:
                if 'fields' not in node:
                    errors.append(ScraperValidationInvalidPluralModifierError(path=path, modifier=modifier))

        # Conflicting leaf keys
        conflicting_leaf_keys = []

        if 'attr' in node:
            conflicting_leaf_keys.append('attr')
        if 'extract' in node:
            conflicting_leaf_keys.append('extract')
        if 'get_context' in node:
            conflicting_leaf_keys.append('get_context')

        if len(conflicting_leaf_keys) > 1:
            errors.append(ScraperValidationConflictError(path=path, keys=conflicting_leaf_keys))

        # Conflicting burrowing/leaf
        if any(k in node for k in BURROWING_KEYS) and any(k in node for k in LEAF_KEYS):
            errors.append(ScraperValidationMixedConcernError(path=path))

        for k, v in node.items():
            p = path + [k]

            # Validating python syntax
            if k == 'eval' or k.endswith('_eval'):
                if isinstance(v, str):
                    try:
                        ast.parse(v)
                    except SyntaxError as e:
                        validation_error = ScraperEvalSyntaxError(
                            reason=e,
                            expression=v,
                            path=p
                        )

                        errors.append(validation_error)
            else:
                k_eval = '%s_eval' % k

                if k_eval in node:
                    validation_error = ScraperValidationConflictError(
                        path=path,
                        keys=[k, k_eval]
                    )

                    errors.append(validation_error)

            if isinstance(v, dict):
                recurse(v, p)

    recurse(scraper)

    return sorted(errors, key=errors_sorting_key)



# =============================================================================
# Minet Facebook Comments CLI Action
# =============================================================================
#
# Logic of the `fb comments` action.
#
import casanova

from minet.constants import COOKIE_BROWSERS
from minet.cli.utils import die, LoadingBar
from minet.facebook import FacebookMobileScraper
from minet.facebook.constants import FACEBOOK_COMMENT_CSV_HEADERS
from minet.facebook.exceptions import (
    FacebookInvalidCookieError,
    FacebookInvalidTargetError
)


def facebook_comments_action(cli_args):
    try:
        scraper = FacebookMobileScraper(cli_args.cookie, throttle=cli_args.throttle)
    except FacebookInvalidCookieError:
        if cli_args.cookie in COOKIE_BROWSERS:
            die([
                'Could not extract relevant cookie from "%s".' % cli_args.cookie
            ])

        die([
            'Relevant cookie not found.',
            'A Facebook authentication cookie is necessary to be able to scrape Facebook comments.',
            'Use the --cookie flag to choose a browser from which to extract the cookie or give your cookie directly.'
        ])

    # Enricher
    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=FACEBOOK_COMMENT_CSV_HEADERS
    )

    # Loading bar
    loading_bar = LoadingBar(
        desc='Scraping comments',
        unit='comment'
    )

    for i, (row, url) in enumerate(enricher.cells(cli_args.column, with_rows=True), 1):
        try:
            batches = scraper.comments(
                url,
                per_call=True,
                detailed=True
            )
        except FacebookInvalidTargetError:
            loading_bar.print('Given url (line %i) is probably not a Facebook resource having comments: %s' % (i, url))
            continue

        for details, batch in batches:
            for comment in batch:
                enricher.writerow(row, comment.as_csv_row())

            loading_bar.update(len(batch))
            loading_bar.update_stats(
                calls=details['calls'],
                replies=details['replies'],
                q=details['queue_size'],
                posts=i
            )



# =============================================================================
# Minet Fetch CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and fetching the urls found
# in the given column. This is done in a respectful multithreaded fashion to
# optimize both running time & memory.
#
import os
import gzip
import casanova
from io import StringIO
from os.path import join, dirname
from collections import Counter
from uuid import uuid4
from ural import is_shortened_url

from minet.fetch import multithreaded_fetch, multithreaded_resolve
from minet.utils import PseudoFStringFormatter
from minet.fs import FolderStrategy
from minet.web import (
    grab_cookies,
    parse_http_header
)
from minet.exceptions import InvalidURLError
from minet.cli.constants import DEFAULT_PREBUFFER_BYTES
from minet.cli.reporters import report_error
from minet.cli.utils import LoadingBar, die


FETCH_ADDITIONAL_HEADERS = [
    'resolved',
    'status',
    'error',
    'filename',
    'mimetype',
    'encoding'
]

RESOLVE_ADDITIONAL_HEADERS = [
    'resolved',
    'status',
    'error',
    'redirects',
    'chain'
]

CUSTOM_FORMATTER = PseudoFStringFormatter()


def fetch_action(cli_args, resolve=False):

    # If we are hitting a single url we enable contents_in_report by default
    if not resolve and isinstance(cli_args.file, StringIO) and cli_args.contents_in_report is None:
        cli_args.contents_in_report = True

    # Trying to instantiate the folder strategy
    if not resolve:
        folder_strategy = FolderStrategy.from_name(cli_args.folder_strategy)

        if folder_strategy is None:
            die([
                'Invalid "%s" --folder-strategy!' % cli_args.folder_strategy,
                'Check the list at the end of the command help:',
                '  $ minet fetch -h'
            ])

    # HTTP method
    http_method = cli_args.method

    # Cookie grabber
    get_cookie = None
    if cli_args.grab_cookies:
        get_cookie = grab_cookies(cli_args.grab_cookies)

    # Global headers
    global_headers = None
    if cli_args.headers:
        global_headers = {}

        for header in cli_args.headers:
            k, v = parse_http_header(header)
            global_headers[k] = v

    # Resume listener
    skipped_rows = 0
    resuming_reader_loading = None

    if cli_args.resume and cli_args.output.can_resume():
        resuming_reader_loading = LoadingBar(
            desc='Resuming',
            unit='line'
        )

        def output_read_listener(event, row):
            nonlocal skipped_rows

            if event != 'output.row':
                return

            skipped_rows += 1
            resuming_reader_loading.update()

        cli_args.output.listener = output_read_listener

    if resolve:
        additional_headers = RESOLVE_ADDITIONAL_HEADERS
    else:
        additional_headers = FETCH_ADDITIONAL_HEADERS

        if cli_args.contents_in_report:
            additional_headers = additional_headers + ['raw_contents']

    # Enricher
    enricher = casanova.threadsafe_enricher(
        cli_args.file,
        cli_args.output,
        add=additional_headers,
        keep=cli_args.select,
        total=cli_args.total,
        prebuffer_bytes=DEFAULT_PREBUFFER_BYTES
    )

    if resuming_reader_loading is not None:
        resuming_reader_loading.close()

    if cli_args.column not in enricher.pos:
        die([
            'Could not find the "%s" column containing the urls in the given CSV file.' % cli_args.column
        ])

    url_pos = enricher.pos[cli_args.column]

    filename_pos = None

    if not resolve and cli_args.filename is not None:
        if cli_args.filename not in enricher.pos:
            die([
                'Could not find the "%s" column containing the filenames in the given CSV file.' % cli_args.filename
            ])

        filename_pos = enricher.pos[cli_args.filename]

    # Loading bar
    loading_bar = LoadingBar(
        desc='Fetching pages',
        total=enricher.total,
        unit='url',
        initial=skipped_rows
    )

    def update_loading_bar(result):
        nonlocal errors

        if result.error is not None:
            errors += 1
        else:
            if resolve:
                status = result.stack[-1].status
            else:
                status = result.response.status

            if status >= 400:
                status_codes[status] += 1

        stats = {'errors': errors}

        for code, count in status_codes.most_common(1):
            stats[str(code)] = count

        loading_bar.update_stats(**stats)
        loading_bar.update()

    only_shortened = getattr(cli_args, 'only_shortened', False)

    def url_key(item):
        url = item[1][url_pos].strip()

        if not url:
            return

        if only_shortened and not is_shortened_url(url):
            return

        # Url templating
        if cli_args.url_template:
            return cli_args.url_template.format(value=url)

        return url

    def request_args(url, item):
        cookie = None

        # Cookie
        if get_cookie:
            cookie = get_cookie(url)

        # Headers
        headers = None

        if global_headers:
            headers = global_headers

        return {
            'method': http_method,
            'cookie': cookie,
            'headers': headers
        }

    def write_fetch_output(index, row, resolved=None, status=None, error=None,
                           filename=None, encoding=None, mimetype=None, data=None):

        addendum = [
            resolved or '',
            status or '',
            error or '',
            filename or '',
            mimetype or '',
            encoding or ''
        ]

        if cli_args.contents_in_report:
            addendum.append(data or '')

        enricher.writerow(index, row, addendum)

    def write_resolve_output(index, row, resolved=None, status=None, error=None,
                             redirects=None, chain=None):
        addendum = [
            resolved or '',
            status or '',
            error or '',
            redirects or '',
            chain or ''
        ]

        enricher.writerow(index, row, addendum)

    errors = 0
    status_codes = Counter()

    common_kwargs = {
        'key': url_key,
        'insecure': cli_args.insecure,
        'threads': cli_args.threads,
        'throttle': cli_args.throttle,
        'domain_parallelism': cli_args.domain_parallelism,
        'max_redirects': cli_args.max_redirects,
        'wait': False,
        'daemonic': True
    }

    if cli_args.timeout is not None:
        common_kwargs['timeout'] = cli_args.timeout

    # Normal fetch
    if not resolve:

        multithreaded_iterator = multithreaded_fetch(
            enricher,
            request_args=request_args,
            **common_kwargs
        )

        for result in multithreaded_iterator:
            index, row = result.item

            if not result.url:

                write_fetch_output(
                    index,
                    row
                )

                loading_bar.update()
                continue

            response = result.response
            data = response.data if response is not None else None

            content_write_flag = 'wb'

            # Updating stats
            update_loading_bar(result)

            # No error
            if result.error is None:

                # Final url target
                resolved_url = response.geturl()

                if resolved_url == result.url:
                    resolved_url = None

                # Should we keep downloaded content?
                if not cli_args.keep_failed_contents and response.status != 200:
                    write_fetch_output(
                        index,
                        row,
                        resolved=resolved_url,
                        status=response.status
                    )
                    continue

                filename = None

                # Building filename
                if data:
                    if filename_pos is not None or cli_args.filename_template:
                        root, ext = os.path.splitext(row[filename_pos]) if filename_pos is not None else (None, '')
                        ext = ext if ext else (result.meta.get('ext') or '')
                        if cli_args.filename_template:
                            filename = CUSTOM_FORMATTER.format(
                                cli_args.filename_template,
                                value=root,
                                ext=ext,
                                line=enricher.wrap(row)
                            )
                        else:
                            filename = root + ext
                    else:
                        # NOTE: it would be nice to have an id that can be sorted by time
                        filename = str(uuid4()) + (result.meta.get('ext') or '')

                    # Applying folder strategy
                    filename = folder_strategy.apply(
                        filename=filename,
                        url=result.response.geturl()
                    )

                # Standardize encoding?
                encoding = result.meta['encoding']

                if data and cli_args.standardize_encoding or cli_args.contents_in_report:
                    if encoding is None or encoding != 'utf-8' or cli_args.contents_in_report:
                        data = data.decode(encoding if encoding is not None else 'utf-8', errors='replace')
                        encoding = 'utf-8'
                        content_write_flag = 'w'

                # Writing file on disk
                if data and not cli_args.contents_in_report:

                    if cli_args.compress:
                        filename += '.gz'

                    resource_path = join(cli_args.output_dir, filename)
                    resource_dir = dirname(resource_path)

                    os.makedirs(resource_dir, exist_ok=True)

                    with open(resource_path, content_write_flag) as f:

                        # TODO: what if standardize_encoding + compress?
                        f.write(gzip.compress(data) if cli_args.compress else data)

                # Reporting in output
                write_fetch_output(
                    index,
                    row,
                    resolved=resolved_url,
                    status=response.status,
                    filename=filename,
                    encoding=encoding,
                    mimetype=result.meta['mimetype'],
                    data=data
                )

            # Handling potential errors
            else:
                error_code = report_error(result.error)

                resolved = None

                if isinstance(result.error, InvalidURLError):
                    resolved = result.error.url

                write_fetch_output(
                    index,
                    row,
                    error=error_code,
                    resolved=resolved
                )

    # Resolve
    else:

        multithreaded_iterator = multithreaded_resolve(
            enricher,
            resolve_args=request_args,
            follow_meta_refresh=cli_args.follow_meta_refresh,
            follow_js_relocation=cli_args.follow_js_relocation,
            infer_redirection=cli_args.infer_redirection,
            **common_kwargs
        )

        for result in multithreaded_iterator:
            index, row = result.item

            if not result.url:

                write_resolve_output(
                    index,
                    row
                )

                loading_bar.update()
                continue

            # Updating stats
            update_loading_bar(result)

            # No error
            if result.error is None:

                # Reporting in output
                last = result.stack[-1]

                write_resolve_output(
                    index,
                    row,
                    resolved=last.url,
                    status=last.status,
                    redirects=len(result.stack) - 1,
                    chain='|'.join(step.type for step in result.stack)
                )

            # Handling potential errors
            else:
                error_code = report_error(result.error)

                write_resolve_output(
                    index,
                    row,
                    error=error_code,
                    redirects=(len(result.stack) - 1) if result.stack else None,
                    chain='|'.join(step.type for step in result.stack) if result.stack else None
                )
